from rest_framework import status
from rest_framework.exceptions import APIException


class CustomBadRequestException(APIException):
    status_code = status.HTTP_400_BAD_REQUEST
    default_detail = "Custom Bad Request Exception"


class CustomConflictException(APIException):
    status_code = status.HTTP_409_CONFLICT
    default_detail = "Conflict has occured"



from typing import OrderedDict

from rest_framework.exceptions import (
    NotFound,
    ValidationError,
)
from rest_framework.test import DjangoRequestFactory

from django.core import exceptions as django_exceptions
from django.db.models import fields
from django.test import TestCase

from audoma.drf.mixins import (  # BulkUpdateModelMixin,
    ActionModelMixin,
    BulkCreateModelMixin,
    BulkUpdateModelMixin,
    CreateModelMixin,
    DestroyModelMixin,
    ListModelMixin,
    RetrieveModelMixin,
    UpdateModelMixin,
)
from audoma.drf.serializers import (
    BulkListSerializer,
    BulkSerializerMixin,
    ModelSerializer,
)
from audoma.drf.viewsets import GenericViewSet
from audoma.tests.testtools import (
    create_basic_view,
    create_model_class,
    create_model_serializer_class,
)


class CommonMixinTestCase(TestCase):
    view_baseclasses = None

    def setUp(self):
        fields_config = {
            "name": fields.CharField(max_length=255),
            "age": fields.IntegerField(),
        }
        self.model = create_model_class(fields_config=fields_config)
        self.serializer_class = create_model_serializer_class(
            meta_model=self.model, meta_fields=["name", "age"]
        )
        self.view = create_basic_view(
            view_baseclasses=self.view_baseclasses,
            view_properties={"serializer_class": self.serializer_class},
        )
        self.view.format_kwarg = "json"
        self.factory = DjangoRequestFactory()


class ActionModelMixinTestCase(CommonMixinTestCase):

    view_baseclasses = (ActionModelMixin, GenericViewSet)

    def test_put_perform_action_with_instance_success(self):
        data = {"name": "John", "age": 47}
        instance = self.model(**data)
        request = self.factory.put("/example")
        request.data = data
        self.view.action = "create"
        self.view.request = request
        response = self.view.perform_action(
            request, instance=instance, success_status=201
        )
        self.assertEqual(response.status_code, 201)
        self.assertEqual(response.data, data)

    def test_post_perform_action_without_instance(self):
        data = {"name": "John", "age": 47}
        request = self.factory.post("/example")
        request.data = data
        self.view.action = "create"
        self.view.request = request
        response = self.view.perform_action(request, success_status=201)
        self.assertEqual(response.status_code, 201)
        self.assertEqual(response.data, data)

    def test_retrieve_instance_success(self):
        def get_object():
            return self.model(name="name", age=21)

        self.view.get_object = get_object
        request = self.factory.post("/example")
        self.view.action = "retrieve"
        self.view.request = request
        response = self.view.retrieve_instance(request)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data, {"name": "name", "age": 21})

    def test_retrieve_instance_failure(self):
        def get_object():
            return

        self.view.get_object = get_object
        request = self.factory.post("/example")
        self.view.action = "retrieve"
        self.view.request = request
        try:
            self.view.retrieve_instance(request)
        except Exception as e:
            self.assertIsInstance(e, AssertionError)

    def test_get_success_headers_success(self):
        headers = self.view.get_success_headers(data={"url": "/header/test/"})
        self.assertEqual(headers, {"Location": "/header/test/"})

    def test_get_success_headers_failure(self):
        try:
            self.view.get_success_headers({})
        except Exception as e:
            self.assertIsInstance(e, KeyError)


class CreateModelMixinTestCase(CommonMixinTestCase):

    view_baseclasses = (CreateModelMixin, GenericViewSet)

    def test_create_success(self):
        data = {"name": "John", "age": 47}
        request = self.factory.post("/example")
        request.data = data
        self.view.action = "create"
        self.view.request = request
        response = self.view.create(request, success_status=201)
        self.assertEqual(response.status_code, 201)
        self.assertEqual(response.data, data)

    def test_create_failure_missing_data(self):
        data = {
            "name": "John",
        }
        request = self.factory.post("/example")
        request.data = data
        self.view.action = "create"
        self.view.request = request
        try:
            self.view.create(request, success_status=201)
        except Exception as e:
            self.assertIsInstance(e, ValidationError)
            self.assertEqual(e.detail["age"][0].code, "required")

    def test_create_failure_incorrect_data(self):
        data = {"name": "Test", "age": "THISISNOTANUMBER"}
        request = self.factory.post("/example")
        request.data = data
        self.view.action = "create"
        self.view.request = request
        try:
            self.view.create(request, success_status=201)
        except Exception as e:
            self.assertIsInstance(e, ValidationError)
            self.assertEqual(e.detail["age"][0].code, "invalid")


class ListModelTestCase(CommonMixinTestCase):
    view_baseclasses = (ListModelMixin, GenericViewSet)

    def setUp(self):
        super().setUp()

        def get_queryset():
            return self.model.objects.none()

        self.view.get_queryset = get_queryset

    def test_list_no_message_success(self):
        request = self.factory.get("/example")
        request.query_params = {"page": 1}
        self.view.action = "list"
        self.view.request = request
        response = self.view.list(request)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data["count"], 0)
        self.assertEqual(response.data["next"], None)
        self.assertEqual(response.data["previous"], None)
        self.assertEqual(response.data["results"], [])
        self.assertEqual(response.data["message"], None)

    def test_list_with_message_success(self):
        def get_list_message():
            return "Important list message"

        self.view.get_list_message = get_list_message

        request = self.factory.get("/example")
        request.query_params = {"page": 1}
        self.view.action = "list"
        self.view.request = request
        response = self.view.list(request)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data["count"], 0)
        self.assertEqual(response.data["next"], None)
        self.assertEqual(response.data["previous"], None)
        self.assertEqual(response.data["results"], [])
        self.assertEqual(response.data["message"], "Important list message")

    def test_list_with_none_page_failure(self):
        request = self.factory.get("/example")
        request.query_params = {"page": None}
        self.view.action = "list"
        self.view.request = request
        try:
            self.view.list(request)
        except Exception as e:
            self.assertEqual(type(e), NotFound)
            self.assertEqual(e.detail, "Invalid page.")


class RetrieveModelMixinTestCase(CommonMixinTestCase):

    view_baseclasses = (RetrieveModelMixin, GenericViewSet)

    def setUp(self):
        super().setUp()

        def get_object():
            return self.model(name="name", age=21)

        self.view.get_object = get_object

    def test_retrieve_success(self):
        request = self.factory.post("/example")
        self.view.action = "retrieve"
        self.view.request = request
        response = self.view.retrieve(request)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data, {"name": "name", "age": 21})


class UpdateModelMixinTestCase(CommonMixinTestCase):
    view_baseclasses = (UpdateModelMixin, GenericViewSet)

    def setUp(self):
        super().setUp()

        def get_object():
            return self.model(name="name", age=21)

        self.view.get_object = get_object

    def test_update_full_success(self):
        data = {"name": "John", "age": 47}
        request = self.factory.put("/example")
        request.data = data
        self.view.request = request
        self.view.action = "update"
        response = self.view.update(request, partial=False)
        response_data = response.data
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response_data["name"], "John")
        self.assertEqual(response_data["age"], 47)

    def test_update_partial_success(self):
        data = {"age": 47}
        request = self.factory.patch("/example")
        request.data = data
        self.view.request = request
        self.view.action = "update"
        response = self.view.update(request, partial=True)
        response_data = response.data
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response_data["age"], 47)


class DestroyModelMixinTestCase(CommonMixinTestCase):

    view_baseclasses = (DestroyModelMixin, GenericViewSet)

    def setUp(self):
        super().setUp()

        def get_object():
            instance = self.model(id=1, name="name", age=21)

            def delete():
                return True

            instance.delete = delete
            return instance

        self.view.get_object = get_object

    def test_destroy_success(self):
        request = self.factory.delete("/example")
        request.data = {"id": 1}
        self.view.request = request
        self.view.action = "destroy"
        response = self.view.destroy(request)
        self.assertEqual(response.status_code, 204)

    def test_destroy_validation_error(self):
        request = self.factory.delete("/example")
        request.data = {"id": 1}

        def perform_destroy(me):
            raise django_exceptions.ValidationError("Some Validation error")

        self.view.perform_destroy = perform_destroy
        self.view.request = request
        self.view.action = "destroy"
        try:
            self.view.destroy(request)
        except Exception as e:
            self.assertEqual(type(e), ValidationError)
            self.assertEqual(str(e.detail["detail"]), "Some Validation error")


class BulkCreateModelMixinTestCase(CommonMixinTestCase):

    view_baseclasses = (BulkCreateModelMixin, GenericViewSet)

    def setUp(self):
        super().setUp()

        self.bulk_serializer_class = create_model_serializer_class(
            meta_model=self.model,
            meta_fields=["name", "age"],
            serializer_base_classes=(BulkSerializerMixin, ModelSerializer),
        )

        class ExampleBulkListSerializer(BulkListSerializer):
            def save(me):
                objs = []
                for x, d in enumerate(me.data):
                    objs.append(self.model(id=x, name=d["name"], age=d["age"]))
                me.instance = objs
                return objs

        self.bulk_serializer_class.Meta.list_serializer_class = (
            ExampleBulkListSerializer
        )

        def save(me):
            data = me.data
            me.instance = self.model(id=1, name=data["name"], age=data["age"])
            return me.instance

        self.bulk_serializer_class.save = save

        self.view.serializer_class = self.bulk_serializer_class

    def test_create_not_bulk_success(self):
        data = {"name": "John", "age": 47}
        request = self.factory.post("/example")
        request.data = data
        self.view.action = "create"
        self.view.request = request
        response = self.view.create(request, success_status=201)
        self.assertEqual(response.status_code, 201)
        self.assertEqual(response.data, data)

    def test_create_is_bulk_success(self):
        data = [{"name": "Greg", "age": 22}, {"name": "John", "age": 47}]
        request = self.factory.post("/example")
        request.data = data
        self.view.action = "create"
        self.view.request = request
        response = self.view.create(request, success_status=201)
        self.assertEqual(response.status_code, 201)
        self.assertEqual(len(response.data), 2)
        self.assertEqual(response.data[0], OrderedDict({"name": "Greg", "age": 22}))
        self.assertEqual(response.data[1], OrderedDict({"name": "John", "age": 47}))


class BulkUpdateMixinTestCase(CommonMixinTestCase):
    view_baseclasses = (BulkUpdateModelMixin, GenericViewSet)

    def setUp(self):
        super().setUp()

        def get_queryset():
            return [
                self.model(id=1, name="Greg", age=22),
                self.model(id=2, name="John", age=47),
            ]

        self.view.get_queryset = get_queryset
        self.bulk_serializer_class = create_model_serializer_class(
            meta_model=self.model,
            meta_fields=["id", "name", "age"],
            serializer_base_classes=(BulkSerializerMixin, ModelSerializer),
        )

        class ExampleBulkListSerializer(BulkListSerializer):
            def save(me):
                for _, d in enumerate(me.validated_data):
                    for x, i in enumerate(me.instance):
                        if i.id == d["id"]:
                            for key, value in d.items():
                                setattr(me.instance[x], key, value)

                return me.instance

        self.bulk_serializer_class.Meta.list_serializer_class = (
            ExampleBulkListSerializer
        )

        def save(me):
            data = me.data
            me.instance = self.model(id=data["id"], name=data["name"], age=data["age"])
            return me.instance

        self.bulk_serializer_class.save = save

        self.view.serializer_class = self.bulk_serializer_class

    def test_bulk_update_not_many_success(self):
        data = [{"id": 2, "name": "John", "age": 55}]
        request = self.factory.put("/example")
        request.data = data
        self.view.action = "update"
        self.view.request = request

        def filter_queryset(queryset):
            data = request.data if isinstance(request.data, list) else [request.data]
            ids = [d["id"] for d in data]
            filtered_queryset = []
            for q in queryset:
                if q.id not in ids:
                    continue
                filtered_queryset.append(q)
            return filtered_queryset

        self.view.filter_queryset = filter_queryset

        response = self.view.bulk_update(request)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data, [OrderedDict(data[0])])

    def test_bulk_update_many_success(self):
        data = [
            {"id": 2, "name": "John", "age": 55},
            {"id": 1, "name": "Zenon", "age": 11},
        ]
        request = self.factory.put("/example")
        request.data = data
        self.view.action = "update"
        self.view.request = request

        def filter_queryset(queryset):
            data = request.data if isinstance(request.data, list) else [request.data]
            ids = [d["id"] for d in data]
            filtered_queryset = []
            for q in queryset:
                if q.id not in ids:
                    continue
                filtered_queryset.append(q)
            return filtered_queryset

        self.view.filter_queryset = filter_queryset

        response = self.view.bulk_update(request)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data, [OrderedDict(data[1]), OrderedDict(data[0])])

    def test_partial_bulk_update_not_many_success(self):
        data = [{"id": 2, "name": "Simon"}]
        request = self.factory.put("/example")
        request.data = data
        self.view.action = "update"
        self.view.request = request

        def filter_queryset(queryset):
            data = request.data if isinstance(request.data, list) else [request.data]
            ids = [d["id"] for d in data]
            filtered_queryset = []
            for q in queryset:
                if q.id not in ids:
                    continue
                filtered_queryset.append(q)
            return filtered_queryset

        self.view.filter_queryset = filter_queryset

        response = self.view.bulk_update(request, partial=True)
        self.assertEqual(response.status_code, 200)
        data[0]["age"] = 47
        self.assertEqual(response.data, [OrderedDict(data[0])])

    def test_partial_bulk_update_many_success(self):
        data = [{"id": 2, "name": "Simon"}, {"id": 1, "name": "Caroline"}]
        request = self.factory.put("/example")
        request.data = data
        self.view.action = "update"
        self.view.request = request

        def filter_queryset(queryset):
            data = request.data if isinstance(request.data, list) else [request.data]
            ids = [d["id"] for d in data]
            filtered_queryset = []
            for q in queryset:
                if q.id not in ids:
                    continue
                filtered_queryset.append(q)
            return filtered_queryset

        self.view.filter_queryset = filter_queryset

        response = self.view.bulk_update(request, partial=True)
        self.assertEqual(response.status_code, 200)
        data[0]["age"] = 47
        data[1]["age"] = 22
        self.assertEqual(response.data, [OrderedDict(data[1]), OrderedDict(data[0])])



# Generated by Django 3.2.12 on 2022-07-06 10:13

import django.db.models.deletion
from django.db import (
    migrations,
    models,
)

import audoma.django.db.fields


class Migration(migrations.Migration):

    dependencies = [
        ("audoma_api", "0004_merge_20220620_1037"),
    ]

    operations = [
        migrations.CreateModel(
            name="Country",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", audoma.django.db.fields.CharField(max_length=225)),
            ],
        ),
        migrations.RemoveField(
            model_name="exampletagmodel",
            name="item",
        ),
        migrations.DeleteModel(
            name="ExampleSimpleModel",
        ),
        migrations.DeleteModel(
            name="ExampleTagModel",
        ),
        migrations.AlterField(
            model_name="exampleperson",
            name="birth_country",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to="audoma_api.country"
            ),
        ),
    ]



import random
import string
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Tuple,
    Type,
    Union,
)

from rest_framework.decorators import action
from rest_framework.fields import (
    Field,
    SerializerMethodField,
)
from rest_framework.request import Request
from rest_framework.serializers import (
    BaseSerializer,
    ModelSerializer,
)
from rest_framework.test import APIRequestFactory
from rest_framework.viewsets import ViewSet

from django.db.models import Model

from audoma.decorators import audoma_action
from audoma.drf.mixins import (
    ActionModelMixin,
    ListModelMixin,
    RetrieveModelMixin,
)
from audoma.drf.viewsets import GenericViewSet


_factory = APIRequestFactory()


def create_basic_view_class(
    docstring: str = "Test View Doc",
    view_baseclasses: Iterable[Type] = None,
    view_properties: Dict[str, Any] = None,
) -> Type[ViewSet]:
    view_baseclasses = view_baseclasses or (
        ListModelMixin,
        RetrieveModelMixin,
        GenericViewSet,
        ActionModelMixin,
    )
    view_properties = view_properties or {}

    class ExampleView(*view_baseclasses):
        ...

    for key, prop in view_properties.items():
        setattr(ExampleView, key, prop)

    ExampleView.__doc__ = docstring
    return ExampleView


def create_basic_view(
    docstring: str = "Test View Doc",
    view_baseclasses: Iterable[Type] = None,
    view_properties: Dict[str, Any] = None,
) -> ViewSet:
    return create_basic_view_class(docstring, view_baseclasses, view_properties)()


def _get_random_name(k: int = 10):
    word = "".join(random.choices(string.ascii_uppercase + string.ascii_lowercase, k=k))
    return word


def create_model_class(
    fields_config: Dict[str, Field],
    model_base_classes: Iterable[Type] = None,
):
    model_base_classes = model_base_classes or (Model,)

    class Meta:
        app_label = "audoma_api"

    model_props = {"Meta": Meta, "__module__": "Example"}
    model_props.update(fields_config)
    model_name = _get_random_name()

    model = type(model_name, model_base_classes, model_props)

    return model


def create_serializer_class(
    fields_config: Dict[str, Field],
    serializer_base_classes: Iterable[Type],
    validators: Iterable[Any] = None,
) -> Type[BaseSerializer]:
    validators = validators or []

    class ExampleSerializer(*serializer_base_classes):
        # Mock serializer save to prevent
        # db connection for mocked serializer
        def save(self, **kwargs):
            if isinstance(self, ModelSerializer):
                instance = self.Meta.model()
                for item, value in self.validated_data.items():
                    setattr(instance, item, value)
                self.instance = instance
                return instance
            return self.validated_data

    for field_name, field in fields_config.items():
        setattr(ExampleSerializer, field_name, field)
        ExampleSerializer._declared_fields[field_name] = field
        if isinstance(field, SerializerMethodField):
            name = field.source or f"get_{field_name}"
            setattr(ExampleSerializer, name, lambda x: None)

    ExampleSerializer.validators = validators

    return ExampleSerializer


def create_serializer(
    fields_config: Dict[str, Field],
    serializer_base_classes: Iterable[Type],
    validators: Iterable[Any] = None,
) -> BaseSerializer:
    return create_serializer_class(fields_config, serializer_base_classes, validators)()


def create_model_serializer_class(
    meta_model: Model,
    serializer_base_classes: Iterable[Type] = (ModelSerializer,),
    meta_fields: Union[str, List[str]] = "__all__",
    fields_config: Dict[str, Field] = {},
    validators: Iterable[Any] = None,
):
    serializer = create_serializer_class(
        fields_config, serializer_base_classes, validators
    )

    class Meta:
        model = meta_model
        fields = meta_fields

    setattr(serializer, "Meta", Meta)
    return serializer


def create_view_with_custom_action(
    docstring: str = "Test View Doc",
    view_baseclasses: Iterable[Type] = None,
    view_properties: Dict[str, Any] = None,
    action_name: str = "custom_action",
    action_serializer_class: BaseSerializer = None,
    action_args: Iterable[Any] = None,
    action_kwargs: Dict[str, Any] = None,
):
    view_baseclasses = view_baseclasses or (
        ListModelMixin,
        RetrieveModelMixin,
        GenericViewSet,
        ActionModelMixin,
    )
    view_properties = view_properties or {}
    action_args = action_args or ()
    action_kwargs = action_kwargs or {}

    class ExampleView(*view_baseclasses):
        permissions_classes = []

        @action(name=action_name, *action_args, **action_kwargs)
        def fun(self, *args, **kwargs):
            ...

    for key, prop in view_properties.items():
        setattr(ExampleView, key, prop)

    ExampleView.__doc__ = docstring

    serializer_name = f"{action_name}_serializer_class"
    ExampleView.fun.__name__ = action_name

    if action_serializer_class is not None:
        setattr(ExampleView, serializer_name, action_serializer_class)
    return ExampleView()


def create_view_with_custom_audoma_action(
    docstring: str = "Test View Doc",
    view_baseclasses: Iterable[Type] = None,
    view_properties: Dict[str, Any] = None,
    request: Request = None,
    action_name: str = "custom_action",
    action_serializer_class: BaseSerializer = None,
    action_args: Iterable[Any] = None,
    action_kwargs: Dict[str, Any] = None,
    returnables: Tuple[int, Any] = None,
    raiseable: Union[Exception, Type[Exception]] = None,
    raise_exception: bool = False,
):
    view_baseclasses = view_baseclasses or (
        ListModelMixin,
        RetrieveModelMixin,
        GenericViewSet,
        ActionModelMixin,
    )
    view_properties = view_properties or {}
    action_args = action_args or ()
    action_kwargs = action_kwargs or {}
    returnables = returnables or ({}, 200)
    raiseable = raiseable or Exception
    Request or _factory.get("/test-request/")

    class ExampleView(*view_baseclasses):
        @audoma_action(name=action_name, *action_args, **action_kwargs)
        def fun(self, request, *args, **kwargs):
            if raise_exception:
                raise raiseable
            return returnables

    for key, prop in view_properties.items():
        setattr(ExampleView, key, prop)

    ExampleView.__doc__ = docstring

    serializer_name = f"{action_name}_serializer_class"
    ExampleView.fun.__name__ = action_name
    setattr(ExampleView, action_name, ExampleView.fun)
    if action_serializer_class is not None:
        setattr(ExampleView, serializer_name, action_serializer_class)
    return ExampleView()



from django.test import TestCase

from audoma.choices import make_choices
from audoma.django.db import fields as django_fields
from audoma.django.forms import MoneyField
from audoma.tests.testtools import create_model_class


class CurrencyFieldTestCase(TestCase):
    def create_currency_field_with_default_currency(self):
        field = django_fields.CurrencyField(default="PLN")
        self.assertEqual(field.example, "PLN")

    def create_currency_field_with_currency_choices(self):
        choices = make_choices("CURRENC", ((0, "PLN", "pln"), (1, "USD", "Usd")))

        field = django_fields.CurrencyField(choices=choices.get_choices())
        self.assertEqual(field.example, "PLN")

    def create_currency_field_without_default_currency_or_choices(self):
        field = django_fields.CurrencyField()
        self.assertEqual(field.example, "XYZ")


class MoneyFieldTestCase(TestCase):
    def setUp(self):
        self.field = django_fields.MoneyField(
            verbose_name="Product's price",
            name="price",
            max_digits=4,
        )
        self.model = create_model_class(fields_config={"price": self.field})

    def test_add_currency_field(self):
        self.field.add_currency_field(self.model, "currency")
        self.assertTrue(hasattr(self.model, "price"))
        self.assertIsInstance(self.field._currency_field, django_fields.CurrencyField)

    def test_formfield_no_kwargs(self):
        formfield = self.field.formfield()
        formfield._has_defaults = False
        self.assertIsInstance(formfield, MoneyField)
        self.assertEqual(formfield.decimal_places, self.field.decimal_places)



from django_filters.filters import (
    ChoiceFilter,
    TypedChoiceFilter,
)
from django_filters.rest_framework import DjangoFilterBackend
from drf_spectacular.plumbing import ComponentRegistry
from rest_framework.filters import SearchFilter
from rest_framework.serializers import Serializer
from rest_framework.test import APIRequestFactory

from django.db import models
from django.test import TestCase

from audoma.choices import make_choices
from audoma.drf import fields as audoma_fields
from audoma.drf.filters import DocumentedTypedChoiceFilter
from audoma.openapi import AudomaAutoSchema
from audoma.schema import (
    AudomaDjangoFilterExtension,
    SearchFilterExtension,
)
from audoma.tests.testtools import (
    create_basic_view,
    create_model_class,
    create_serializer_class,
)


class AudomaDjangoFilterExtensionTestCase(TestCase):
    def setUp(self):
        self.factory = APIRequestFactory()
        self.choices = make_choices(
            "RATE", ((0, "LIKE", "Like"), (1, "DISLIKIE", "Dislike"))
        )
        self.fields_config = {
            "company_name": audoma_fields.CharField(max_length=255),
            "city": audoma_fields.CharField(),
            "company_rate": audoma_fields.ChoiceField(
                choices=self.choices.get_choices()
            ),
        }
        serializer_class = create_serializer_class(
            fields_config=self.fields_config,
            serializer_base_classes=[Serializer],
        )
        request = self.factory.post("/example/")
        self.model_fields_config = {
            "company_name": models.CharField(max_length=255),
            "city": models.CharField(max_length=255),
            "company_rate": models.CharField(choices=self.choices.get_choices()),
        }
        model = create_model_class(fields_config=self.model_fields_config)
        self.view = create_basic_view(
            view_properties={"serializer_class": serializer_class}
        )
        self.view.request = request
        self.view.action = "create"
        self.view.schema = AudomaAutoSchema()
        self.view.schema.registry = ComponentRegistry()
        self.view.schema.method = "POST"
        self.view.model = model
        self.extension = AudomaDjangoFilterExtension(target=DjangoFilterBackend())

    def test_get_audoma_documented_typed_choice_filter_x_choices(self):
        field = DocumentedTypedChoiceFilter(
            full_choices=self.choices, parameter_name="company_rate"
        )
        result = self.extension.resolve_filter_field(
            self.view.schema, self.view.model, None, "company_rate", field
        )
        self.assertIsInstance(result, list)
        self.assertEqual(
            result,
            [
                {
                    "in": "query",
                    "name": "company_rate",
                    "schema": {
                        "type": "string",
                        "enum": ["DISLIKIE", "LIKE"],
                        "x-choices": {
                            "choices": {"LIKE": "Like", "DISLIKIE": "Dislike"}
                        },
                    },
                    "description": "Filter by None \n * `LIKE` - Like\n * `DISLIKIE` - Dislike\n",
                }
            ],
        )

    def test_get_typed_choice_filter_x_choices(self):
        field = TypedChoiceFilter(
            choices=self.choices.get_api_choices(), parameter_name="company_rate"
        )
        result = self.extension.resolve_filter_field(
            self.view.schema, self.view.model, None, "company_rate", field
        )
        self.assertEqual(
            result[0]["schema"]["x-choices"],
            {"choices": {"LIKE": "Like", "DISLIKIE": "Dislike"}},
        )

    def test_get_choice_filter_x_choices(self):
        field = ChoiceFilter(
            choices=self.choices.get_api_choices(), parameter_name="company_rate"
        )
        result = self.extension.resolve_filter_field(
            self.view.schema, self.view.model, None, "company_rate", field
        )
        self.assertEqual(
            result[0]["schema"]["x-choices"],
            {"choices": {"LIKE": "Like", "DISLIKIE": "Dislike"}},
        )

    def test_get_x_choices_no_choices(self):
        field = ChoiceFilter()
        result = self.extension.resolve_filter_field(
            self.view.schema, self.view.model, None, "company_rate", field
        )
        self.assertEqual(result, [])


class SearchFilterExtensionTestCase(TestCase):
    def setUp(self):
        self.factory = APIRequestFactory()
        self.choices = make_choices(
            "RATE", ((0, "LIKE", "Like"), (1, "DISLIKIE", "Dislike"))
        )
        self.fields_config = {
            "company_name": audoma_fields.CharField(max_length=255),
            "city": audoma_fields.CharField(),
            "company_rate": audoma_fields.ChoiceField(
                choices=self.choices.get_choices()
            ),
        }
        serializer_class = create_serializer_class(
            fields_config=self.fields_config,
            serializer_base_classes=[Serializer],
        )
        request = self.factory.post("/example/")
        self.model_fields_config = {
            "company_name": models.CharField(max_length=255),
            "city": models.CharField(max_length=255),
            "company_rate": models.CharField(choices=self.choices.get_choices()),
        }
        model = create_model_class(fields_config=self.model_fields_config)
        self.view = create_basic_view(
            view_properties={"serializer_class": serializer_class}
        )
        self.view.request = request
        self.view.action = "create"
        self.view.schema = AudomaAutoSchema()
        self.view.schema.registry = ComponentRegistry()
        self.view.schema.method = "POST"
        self.view.model = model
        self.view.search_fields = ["company_name", "city"]
        self.extension = SearchFilterExtension(target=SearchFilter())

    def test_search_filter_generate_default_description(self):
        result = self.extension.get_schema_operation_parameters(self.view.schema)
        self.assertEqual(
            result[0]["description"], "Search by: \n* `company_name` \n* `city` \n"
        )

    def test_search_filter_generate_custom_simple_description(self):
        self.extension.target.search_description = (
            "Some Custom description.\n {search_fields}"
        )
        result = self.extension.get_schema_operation_parameters(self.view.schema)
        self.assertEqual(
            result[0]["description"],
            "Some Custom description.\n Search by: \n* `company_name` \n* `city` \n",
        )



from rest_framework.permissions import BasePermission


class ViewAndDetailPermission(BasePermission):
    """The ViewAndDetailPermission guard access to all views nad detail views (collections and resources).
    General view permission is determined first then if it's detail view specific object permission is checked.
    """

    def has_permission(self, request, view):
        return True

    def has_object_permission(self, request, view, obj):
        return True


class DetailPermission(BasePermission):
    """The DetailPermission guard access to Detail Views (resources)."""

    def has_object_permission(self, request, view, obj):
        return True


class ViewPermission(BasePermission):
    """The ViewPermission guard access to all views in ViewSet (collections and resources)."""

    def has_permission(self, request, view):
        return True


class AlternatePermission1(BasePermission):
    """The AlternatePermission1 is combined through OR operator with AlternatePermission2"""

    def has_permission(self, request, view):
        return True


class AlternatePermission2(BasePermission):
    """The AlternatePermission2 is combined through OR operator with AlternatePermission1"""

    def has_permission(self, request, view):
        return True


# check_object_permissions



from django.test import TestCase

from audoma.choices import make_choices
from audoma.django.db import fields as db_fields
from audoma.drf import (
    fields,
    serializers,
)
from audoma.tests import testtools


class ResultTestCase(TestCase):
    def setUp(self):
        fields_config = {
            "name": fields.CharField(max_length=255),
            "age": fields.IntegerField(),
        }
        self.serializer_class = testtools.create_serializer_class(
            fields_config=fields_config,
            serializer_base_classes=(serializers.Serializer,),
        )

    def test_result_serializer_class_class_name_endswith_serializer(self):
        self.serializer_class.__name__ = "ExampleDetailSerializer"
        wrapped_serializer = serializers.result_serializer_class(self.serializer_class)
        self.assertEqual(wrapped_serializer.__name__, "ExampleDetailResultSerializer")

    def test_result_serializer_class_different_class_name(self):
        self.serializer_class.__name__ = "Example"
        wrapped_serializer = serializers.result_serializer_class(self.serializer_class)
        self.assertEqual(wrapped_serializer.__name__, "ExampleResult")

    def test_result_serializer_class_many(self):
        self.serializer_class.__name__ = "ExampleDetailSerializer"
        wrapped_serializer = serializers.result_serializer_class(self.serializer_class)
        self.assertTrue(hasattr(wrapped_serializer(many=True).instance, "results"))

    def test_result_serializer_class_not_many(self):
        self.serializer_class.__name__ = "ExampleDetailSerializer"
        wrapped_serializer = serializers.result_serializer_class(self.serializer_class)
        self.assertTrue(hasattr(wrapped_serializer().instance, "result"))


class ResultSerializerClassMixinTestCase(TestCase):
    def setUp(self):
        super().setUp()
        self.fields_config = {
            "name": serializers.CharField(max_length=255),
            "age": serializers.IntegerField(),
        }
        self.base_classes = (serializers.Serializer,)

    def test_wrap_serializer_class_many(self):
        serializer_class = testtools.create_serializer_class(
            fields_config=self.fields_config, serializer_base_classes=self.base_classes
        )
        serializer_class._wrap_result_serializer = True
        new_serializer_class = serializer_class.get_result_serializer_class()
        self.assertNotEqual(serializer_class, new_serializer_class)
        serializer = new_serializer_class(many=True)
        self.assertTrue(hasattr(serializer.instance, "results"))
        self.assertFalse(hasattr(serializer.instance, "result"))

    def test_wrap_serializer_class_not_many(self):

        serializer_class = testtools.create_serializer_class(
            fields_config=self.fields_config, serializer_base_classes=self.base_classes
        )
        serializer_class._wrap_result_serializer = True
        new_serializer_class = serializer_class.get_result_serializer_class()
        self.assertNotEqual(serializer_class, new_serializer_class)
        serializer = new_serializer_class(many=False)
        self.assertTrue(hasattr(serializer.instance, "result"))
        self.assertFalse(hasattr(serializer.instance, "results"))

    def test_no_wrap_serializer_class_many(self):
        serializer_class = testtools.create_serializer_class(
            fields_config=self.fields_config, serializer_base_classes=self.base_classes
        )
        serializer_class._wrap_result_serializer = False
        new_serializer_class = serializer_class.get_result_serializer_class()
        self.assertEqual(new_serializer_class, serializer_class)

    def test_no_wrap_serializer_class_not_many(self):
        serializer_class = testtools.create_serializer_class(
            fields_config=self.fields_config, serializer_base_classes=self.base_classes
        )
        serializer_class._wrap_result_serializer = False
        new_serializer_class = serializer_class.get_result_serializer_class()
        self.assertEqual(new_serializer_class, serializer_class)


class ModelSerializerTestCase(TestCase):
    def test_build_standard_field_with_example(self):
        model_fields_config = {
            "name": db_fields.CharField(max_length=255, example="Thomas"),
            "age": db_fields.IntegerField(example=33),
        }

        class ExampleModelSerializer(serializers.ModelSerializer):
            class Meta:
                model = testtools.create_model_class(fields_config=model_fields_config)
                fields = ["name", "age"]

        serializer = ExampleModelSerializer()
        field, field_kwargs = serializer.build_standard_field(
            "name", model_fields_config["name"]
        )
        self.assertEqual(field, serializers.CharField)
        self.assertEqual(field_kwargs["example"], "Thomas")

        field, field_kwargs = serializer.build_standard_field(
            "age", model_fields_config["age"]
        )
        self.assertEqual(field, serializers.IntegerField)
        self.assertEqual(field_kwargs["example"], 33)

    def test_build_standard_field_with_no_example(self):
        model_fields_config = {
            "name": db_fields.CharField(max_length=255),
            "age": db_fields.IntegerField(),
        }

        class ExampleModelSerializer(serializers.ModelSerializer):
            class Meta:
                model = testtools.create_model_class(fields_config=model_fields_config)
                fields = ["name", "age"]

        serializer = ExampleModelSerializer()
        field, field_kwargs = serializer.build_standard_field(
            "name", model_fields_config["name"]
        )
        self.assertEqual(field, serializers.CharField)
        self.assertIsNotNone(field_kwargs["example"])

        field, field_kwargs = serializer.build_standard_field(
            "age", model_fields_config["age"]
        )


class DisplayNamedWritableFieldTestCase(TestCase):
    def setUp(self):
        self.choices = make_choices("COLOR", ((0, "RED", "Red"), (1, "BLUE", "Blue")))
        self.field = serializers.DisplayNameWritableField(
            choices=self.choices.get_choices()
        )

    def test_to_representation_success(self):
        representation = self.field.to_representation(0)
        self.assertEqual(representation, "Red")
        representation = self.field.to_representation(1)
        self.assertEqual(representation, "Blue")

    def test_to_internal_value_success(self):
        representation = self.field.to_internal_value("Red")
        self.assertEqual(representation, 0)
        representation = self.field.to_internal_value("Blue")
        self.assertEqual(representation, 1)

    def test_to_internal_value_validation_error_failure(self):
        try:
            self.field.to_internal_value("Tom")
        except Exception as e:
            self.assertEqual(type(e), serializers.ValidationError)
            self.assertEqual(str(e.detail[0]), '"Tom" is not valid choice.')
            self.assertEqual(e.detail[0].code, "invalid")



import json
import random
from typing import (
    Any,
    List,
)

from rest_framework import viewsets
from rest_framework.pagination import PageNumberPagination
from rest_framework.response import Response

from audoma.drf.generics import GenericAPIView


class AudomaPagination(PageNumberPagination):
    """
    A simple page number based style that supports page numbers as
    query parameters.

    Note:
        If this won't be used it'll cause less explicit pagination documentation.
        This class does not provide any additional functionality.

    Args:
        page_size (int) - number of items per page - by default this is set to 25
        max_page_size (int) - maximum number of items per page - by default this is set to 2000
    """

    page_size = 25
    max_page_size = 2000

    def get_paginated_response_schema(self, schema: List[dict]) -> dict:
        """
        Simple method to add pagination information to the schema.

        Args:
            schema (List[dict]) - list of schema items
        Returns:
            Dictionary with pagination information including examples
        """
        return {
            "type": "object",
            "properties": {
                "count": {"type": "integer", "example": random.randint(1, 100)},
                "message": {
                    "type": "string",
                    "nullable": True,
                },
                "next": {
                    "type": "string",
                    "nullable": True,
                    "format": "uri",
                    "example": "http://api.example.org/accounts/?page=4",
                },
                "previous": {
                    "type": "string",
                    "nullable": True,
                    "format": "uri",
                    "example": "http://api.example.org/accounts/?page=2",
                },
                "results": schema,
            },
        }


class GenericViewSet(viewsets.ViewSetMixin, GenericAPIView):
    pagination_class = AudomaPagination

    def _parse_response_data(self, response_data: List[Any]) -> List[str]:
        parsed_data = []
        for item in response_data:
            if isinstance(item, str):
                parsed_data.append(item)
            else:
                parsed_data.append(json.dumps(item))
        return parsed_data

    def handle_exception(self, exc: Exception) -> Response:
        response = super().handle_exception(exc)
        if isinstance(response.data, dict):
            if response.status_code != 418:
                for k in response.data:
                    if isinstance(response.data[k], list):
                        if not all(isinstance(item, str) for item in response.data[k]):
                            response.data[k] = self._parse_response_data(
                                response.data[k]
                            )
                        response.data[k] = " ".join(response.data[k])
                response.data = {"errors": response.data}
        return response



from rest_framework import exceptions

from django.conf import settings


WRAP_RESULT_SERIALIZER = getattr(settings, "AUDOMA_WRAP_RESULT_SERIALIZER", False)
settings.SPECTACULAR_SETTINGS[
    "GET_LIB_DOC_EXCLUDES"
] = "audoma.plumbing.get_lib_doc_excludes_audoma"


COMMON_API_ERRORS = [
    exceptions.NotFound,
    exceptions.NotAuthenticated,
    exceptions.AuthenticationFailed,
    exceptions.ParseError,
    exceptions.PermissionDenied,
    exceptions.NotAcceptable,
    exceptions.ValidationError,
    exceptions.Throttled,
]

settings.SPECTACULAR_SETTINGS["POSTPROCESSING_HOOKS"] = [
    "audoma.hooks.postprocess_common_errors_section",
]
